use ahash::AHashMap;
use line_index::LineIndex;
use lineage::{Lineage, Node};
use sqruff_lib::core::config::FluffConfig;
use sqruff_lib::core::linter::core::Linter as SqruffLinter;
use sqruff_lib_core::parser::parser::Parser;
use sqruff_lib_core::parser::segments::base::Tables;
use wasm_bindgen::prelude::*;

#[wasm_bindgen]
#[derive(Clone)]
pub struct Diagnostic {
    message: String,
    pub start_line_number: u32,
    pub start_column: u32,
    pub end_line_number: u32,
    pub end_column: u32,
}

#[wasm_bindgen]
impl Diagnostic {
    #[wasm_bindgen(getter)]
    pub fn message(&self) -> String {
        self.message.clone()
    }
}

#[wasm_bindgen]
pub struct Linter {
    base: SqruffLinter,
}

#[wasm_bindgen]
#[derive(PartialEq, Eq)]
pub enum Tool {
    Format = "Format",
    Cst = "Cst",
    Lineage = "Lineage",
}

#[wasm_bindgen]
#[derive(Default)]
pub struct Result {
    diagnostics: Vec<Diagnostic>,
    secondary: String,
}

#[wasm_bindgen]
impl Result {
    #[wasm_bindgen(constructor)]
    pub fn new() -> Result {
        Result::default()
    }
}

#[wasm_bindgen]
impl Result {
    #[wasm_bindgen(getter)]
    pub fn diagnostics(&self) -> Vec<Diagnostic> {
        self.diagnostics.clone()
    }

    #[wasm_bindgen(getter)]
    pub fn secondary(&self) -> String {
        self.secondary.clone()
    }
}

#[wasm_bindgen]
impl Linter {
    #[wasm_bindgen(constructor)]
    pub fn new(source: &str) -> Self {
        Self {
            base: SqruffLinter::new(FluffConfig::from_source(source, None), None, None, true),
        }
    }

    #[wasm_bindgen]
    pub fn check(&self, sql: &str, tool: Tool) -> Result {
        let line_index = LineIndex::new(sql);

        let tables = Tables::default();
        let parsed = self.base.parse_string(&tables, sql, None).unwrap();

        let cst = if tool == Tool::Cst {
            parsed.tree.clone()
        } else {
            None
        };

        let mut result = self.base.lint_parsed(&tables, parsed, tool == Tool::Format);
        let violations = &mut result.violations;

        let diagnostics = violations
            .iter_mut()
            .map(|violation| {
                let start = line_index.line_col(violation.source_slice.start.try_into().unwrap());
                let end = line_index.line_col(violation.source_slice.end.try_into().unwrap());

                Diagnostic {
                    message: std::mem::take(&mut violation.description),
                    start_line_number: start.line + 1,
                    start_column: start.col + 1,
                    end_line_number: end.line + 1,
                    end_column: end.col + 1,
                }
            })
            .collect();

        let secondary = match tool {
            Tool::Format => result.fix_string(),
            Tool::Cst => cst.unwrap().stringify(false),
            Tool::Lineage => {
                let parser = Parser::new(self.base.config().get_dialect(), AHashMap::new());
                let (tables, node) = Lineage::new(parser, "", sql).build();

                print_tree(&tables, node, "", "", "")
            }
            Tool::__Invalid => String::from("Error: unsupported tool"),
        };

        Result {
            diagnostics,
            secondary,
        }
    }
}

fn print_tree(
    tables: &lineage::ir::Tables,
    node: Node,
    parent_prefix: &str,
    immediate_prefix: &str,
    parent_suffix: &str,
) -> String {
    use std::fmt::Write;

    let node_data = &tables.nodes[node];

    let name = &node_data.name;
    let source = tables.stringify(node_data.source);
    let expression = tables.stringify(node_data.expression);
    let source_name = &node_data.source_name;
    let reference_node_name = &&node_data.reference_node_name;

    let mut string = String::new();

    let _ = writeln!(
        string,
        "{:1$}{parent_prefix}{immediate_prefix}name: {name}",
        "", 0
    );
    let _ = writeln!(
        string,
        "{:1$}{parent_prefix}{immediate_prefix}source: {source}",
        "", 0
    );
    let _ = writeln!(
        string,
        "{:1$}{parent_prefix}{immediate_prefix}expression: {expression}",
        "", 0
    );
    let _ = writeln!(
        string,
        "{:1$}{parent_prefix}{immediate_prefix}source_name: {source_name}",
        "", 0
    );
    let _ = writeln!(
        string,
        "{:1$}{parent_prefix}{immediate_prefix}reference_node_name: {reference_node_name}",
        "", 0
    );

    let mut it = node_data.downstream.iter().peekable();
    let child_prefix = format!("{0}{1}", parent_prefix, parent_suffix);

    while let Some(child) = it.next().copied() {
        let ret = match it.peek() {
            None => print_tree(tables, child, &child_prefix, "└─ ", "   "),
            Some(_) => print_tree(tables, child, &child_prefix, "├─ ", "│  "),
        };

        string.push_str(&ret);
    }

    string
}
